#include <iostream> // Stream-basierte Ein-/Ausgabe kommt noch
#include <string> // String"klasse" kommt noch

using namespace std; // wegen neuer Standard-Header (die ohne .h) ntig
                     // Dev-C++ (also mingw) kommt auch ohne aus.

#define AUSGABE(x) cout << (x) << endl; // Dieses bitte erst mal ignorieren.

class TestA
{
    private:

    int mgroesse;
    int* p;

    protected:

    string mname;

    public:

    TestA (string name = "NN", int groesse = 1) :
    mgroesse (groesse),  // schnere Initalisierung. Achtung: Komma, kein Semikolon!
    mname (name)
    {
        p = new int[groesse];
        AUSGABE("TestA "+mname+" wird konstruiert.")
    }

    /* virtual */ ~TestA ()
    {
        delete[] p;
        AUSGABE("TestA "+mname+" wird destruiert.")
    }

    /* virtual */ void setze (int wo, int wert = 0)
    {
        p[wo] = wert;
        AUSGABE("\"setze\" fuer TestA in "+mname)
    }

    /* virtual */ int lese (int wo)
    {
        AUSGABE("\"lese\" fuer TestA in "+mname)
        return p[wo];
    }
};

class TestB : public TestA    // in Java: extends TestA
{
    private:

    int zahlDerZugriffe;

    public:

    TestB (string name = "NN", int groesse = 1) :
    TestA (name, groesse),
    zahlDerZugriffe (0)  // Lies:  zahlDerZugriffe = 0;
    {
        AUSGABE("TestB "+mname+" wird konstruiert.")
    }

    ~TestB ()
    {
        AUSGABE("TestB "+mname+" wird destruiert.")
    }  // Destruktor von TestA wird automatisch aufgerufen,
       // das reicht hier.

    void setze (int wo, int wert = 0) // Die Funktion setze von TestA
                                      // wird "berschrieben", d.h. in der
                                      // abgeleiteten Klasse durch eine
                                      // neue Definition ersetzt.
    {
        zahlDerZugriffe++;
        AUSGABE("\"setze\" fuer TestB in "+mname)
        TestA::setze(wo,wert);  // Methode der Mutterklasse. In Java mittels "super".
    }

    int lese (int wo)
    {
        zahlDerZugriffe++;
        AUSGABE("\"lese\" fuer TestB in "+mname)
        return TestA::lese(wo);
    }
};

// "Polymorphie" = Vielgestaltigkeit,
// d.h. abgeleitete Klassen lassen sich ber Pointer und Referenzen ansprechen,
// die den Typ der Mutterklasse haben. Alter oder allgemeiner Programmcode
// kann damit auf neuen, spezialisierten Klassen arbeiten.

int main ( int argc, char* (argv[]) )
{
    AUSGABE("-----------------------------1")
    int i;
    for (i=0; i<1; i++)  // als Beispiel fr Blockbildung
    {
        TestB abc("abc",13);
        abc.setze(10,12);
    }
    AUSGABE("-----------------------------2")
    for (i=0; i<1; i++)  // als Beispiel fr Blockbildung
    {
        TestB bcd("bcd",13);
        TestA* pa = &bcd;
        //! TestA a = bcd; // Geht nicht sinnvoll: bcd wrde kleingehobelt, weil
                           // nur Speicher fr ein TestA-Objekt da ist.
        pa->setze(10,12);  // Dumm: Ruft die Funktion "setze" von TestA auf.
                           // Java wrde die von TestB aufrufen! ("late binding")
                           // Oben die Kommentare um "virtual" weg,
                           // dann gehts auch mit C++ so. ("virtuelle Methoden")
    }
    AUSGABE("-----------------------------3")
    for (i=0; i<1; i++)  // als Beispiel fr Blockbildung
    {
        TestB cde("cde",13);
        TestA& a = cde;
        a.setze(10,12);  // Bei Referenzen genau dasselbe.
    }
    AUSGABE("-----------------------------4")
    for (i=0; i<1; i++)  // als Beispiel fr Blockbildung
    {
        TestA* pa = new TestB ("def", 13);
        pa->setze(10,12);
        delete pa;        // GEFAHR! Ruft nur den Destruktor von TestA auf,
                          // wenn der nicht als "virtual" deklariert ist.
    }
    return 0;
}
