#include <iostream> // Stream-basierte Ein-/Ausgabe kommt noch
#include <string> // String"klasse" kommt noch

using namespace std; // wegen neuer Standard-Header (die ohne .h) ntig
                     // Dev-C++ (also mingw) kommt auch ohne aus.

#define AUSGABE(x) cout << (x) << endl; // Dieses bitte erst mal ignorieren.

// "Default-Methoden"
// Der Compiler erzeugt automatisch folgende drei Methoden, wenn man sie
// nicht selbst vorgibt:
// * Konstruktor (wird dann einer ohne Parameter)
// * Zuweisungsoperator
// * Copy-Konstruktur
// Sobald dynamisch angeforderter Speicher im Spiel ist, muss man diese
// Methoden meist selbst definieren. Und natrlich einen Destruktor.

// Konstruktoraufrufe:
//    Auto Trabbi("HB-XY-1234");
//    Auto Trabbi = Auto("HB-XY-1234");  // Auch wenn es wie Zuweisung aussieht.
//    Auto* p = new Auto("HB-XY-1234");

// Aufrufe des Zuweisungsoperators:
//    Auto Trabbi;
//    Trabbi = Auto("HB-XY-1234");  // Hier!

// Aufrufe des Copy-Konstruktors:
//    Auto Trabbi("HB-XY-1234");
//    Auto a = Trabbi;    // Hier! Auch wenns wie Zuweisung aussieht.
//    oder:
//    Auto lackieren (Auto a)  // Hier!
//    {
//        return a;  // Hier!
//    }



// Simples Beispiel:

class ATest
{
    private:
    int* p; // soll nachher auf einen Speicherbereich zeigen,
            // in dem das Objekt seine Daten ablegt
};

// Daraus entsteht im Complier automatisch:

class BTest
{
    private:
    int* p;

    public:

    BTest(void) {} // So she der Default-Konstruktor aus,
                   // ***WENN*** man ihn schreiben wrde.
                   // Ruft implizit Konstruktoren aller Mutter- und
                   // Unterobjekte auf.

    const BTest& operator= (const BTest& x) // So she der Default-Zuweisungsoperator aus.
    {                                       // Ruft implizit die Zuweisungsoperatoren enthaltener Objekte auf.
        p = x.p; // fr elementare Datentypen bitweise (=flache) Kopie
        return *this;
    }

    BTest(const BTest& x) // So she der Default-Copy-Konstruktor aus.
    {                     // Ruft implizit die Copy-Konstruktoren enthaltener Objekte auf.
        p = x.p; // fr elementare Datentypen bitweise (=flache) Kopie
    }
};

// Sinnvoll wre dagegen:

class CTest
{
    private:
    int* p;

    public:

    CTest(void)
    {
        AUSGABE("Konstruktor")
        p = new int[1000]; // Speicher reservieren!
    }

    const CTest& operator= (const CTest& x)
    {
        AUSGABE("Zuweisungsoperator")
        for (int i=0; i<1000; i++)
        {
            p[i] = (x.p)[i]; // tiefe Kopie
        }
        return *this;
    }

    CTest(const CTest& x)
    {
        AUSGABE("Copy-Konstruktor")
        p = new int[1000]; // Speicher reservieren, ist ja ein Konstruktor!
        for (int i=0; i<1000; i++)
        {
            p[i] = (x.p)[i]; // tiefe Kopie
        }
    }

    ~CTest() // Wir brauchen einen Destruktor, ...
    {
        delete[] p; // ... um den Speicher freizugeben.
    }
};

CTest test (CTest a)
{
    return a;
}

int main (int argc, char* (argv[]))
{
    CTest a;
    AUSGABE("-----------------------------------------")
    a = CTest();
    AUSGABE("-----------------------------------------")
    CTest b = a;
    AUSGABE("-----------------------------------------")
    a = test(b);
    return 0;
}
